"""
Example generation for Py-ART

Generate the rst files for the examples by iterating over the Python
example files.

Files that generate images should state with 'plot'

"""

# This files was adapted from the scikit-learn file of the same name:
# https://github.com/scikit-learn/scikit-learn/blob/master/doc/sphinxext/gen_rst.py
#
# The scikit-learn project uses this following License which applies to this
# file.
#
# New BSD License
#
# Copyright (c) 2007-2013 The scikit-learn developers.
# All rights reserved.
#
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the Scikit-learn Developers nor the names of
#      its contributors may be used to endorse or promote products
#      derived from this software without specific prior written
#      permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

from __future__ import division, print_function

from time import time
import os
import re
import shutil
import traceback
import glob
import sys
import gzip
import posixpath

# Try Python 2 first, otherwise load from Python 3
try:
    from StringIO import StringIO
    import cPickle as pickle
    import urllib2 as urllib
    from urllib2 import HTTPError, URLError
except:
    from io import StringIO
    import pickle
    import urllib.request
    import urllib.error
    import urllib.parse
    from urllib.error import HTTPError, URLError

try:
    # Python 2 built-in
    execfile
except NameError:
    def execfile(filename, global_vars=None, local_vars=None):
        with open(filename, encoding='utf-8') as f:
            code = compile(f.read(), filename, 'exec')
            exec(code, global_vars, local_vars)

try:
    basestring
except NameError:
    basestring = str

try:
    from PIL import Image
except:
    import Image

import matplotlib
matplotlib.use('Agg')

import token
import tokenize
import numpy as np


###############################################################################
# A tee object to redict streams to multiple outputs

class Tee(object):

    def __init__(self, file1, file2):
        self.file1 = file1
        self.file2 = file2

    def write(self, data):
        self.file1.write(data)
        self.file2.write(data)

    def flush(self):
        self.file1.flush()
        self.file2.flush()

###############################################################################
# Documentation link resolver objects


def get_data(url):
    """Helper function to get data over http or from a local file"""
    if url.startswith('http://'):
        resp = urllib.urlopen(url)
        encoding = resp.headers.dict.get('content-encoding', 'plain')
        data = resp.read()
        if encoding == 'plain':
            pass
        elif encoding == 'gzip':
            data = StringIO(data)
            data = gzip.GzipFile(fileobj=data).read()
        else:
            raise RuntimeError('unknown encoding')
    else:
        with open(url, 'r') as fid:
            data = fid.read()
        fid.close()

    return data


def parse_sphinx_searchindex(searchindex):
    """Parse a Sphinx search index

    Parameters
    ----------
    searchindex : str
        The Sphinx search index (contents of searchindex.js)

    Returns
    -------
    filenames : list of str
        The file names parsed from the search index.
    objects : dict
        The objects parsed from the search index.
    """
    def _select_block(str_in, start_tag, end_tag):
        """Select first block delimited by start_tag and end_tag"""
        start_pos = str_in.find(start_tag)
        if start_pos < 0:
            raise ValueError('start_tag not found')
        depth = 0
        for pos in range(start_pos, len(str_in)):
            if str_in[pos] == start_tag:
                depth += 1
            elif str_in[pos] == end_tag:
                depth -= 1

            if depth == 0:
                break
        sel = str_in[start_pos + 1:pos]
        return sel

    def _parse_dict_recursive(dict_str):
        """Parse a dictionary from the search index"""
        dict_out = dict()
        pos_last = 0
        pos = dict_str.find(':')
        while pos >= 0:
            key = dict_str[pos_last:pos]
            if dict_str[pos + 1] == '[':
                # value is a list
                pos_tmp = dict_str.find(']', pos + 1)
                if pos_tmp < 0:
                    raise RuntimeError('error when parsing dict')
                value = dict_str[pos + 2: pos_tmp].split(',')
                # try to convert elements to int
                for i in range(len(value)):
                    try:
                        value[i] = int(value[i])
                    except ValueError:
                        pass
            elif dict_str[pos + 1] == '{':
                # value is another dictionary
                subdict_str = _select_block(dict_str[pos:], '{', '}')
                value = _parse_dict_recursive(subdict_str)
                pos_tmp = pos + len(subdict_str)
            else:
                raise ValueError('error when parsing dict: unknown elem')

            key = key.strip('"')
            if len(key) > 0:
                dict_out[key] = value

            pos_last = dict_str.find(',', pos_tmp)
            if pos_last < 0:
                break
            pos_last += 1
            pos = dict_str.find(':', pos_last)

        return dict_out

    # parse objects
    query = 'objects:'
    pos = searchindex.find(query)
    if pos < 0:
        raise ValueError('"objects:" not found in search index')

    sel = _select_block(searchindex[pos:], '{', '}')
    objects = _parse_dict_recursive(sel)

    # parse filenames
    query = 'filenames:'
    pos = searchindex.find(query)
    if pos < 0:
        raise ValueError('"filenames:" not found in search index')
    filenames = searchindex[pos + len(query) + 1:]
    filenames = filenames[:filenames.find(']')]
    filenames = [f.strip('"') for f in filenames.split(',')]

    return filenames, objects


class SphinxDocLinkResolver(object):
    """ Resolve documentation links using searchindex.js generated by Sphinx

    Parameters
    ----------
    doc_url : str
        The base URL of the project website.
    searchindex : str
        Filename of searchindex, relative to doc_url.
    extra_modules_test : list of str
        List of extra module names to test.
    relative : bool
        Return relative links (only useful for links to documentation of this
        package).
    """

    def __init__(self, doc_url, searchindex='searchindex.js',
                 extra_modules_test=None, relative=False):
        self.doc_url = doc_url
        self.relative = relative
        self._link_cache = {}

        self.extra_modules_test = extra_modules_test
        self._page_cache = {}
        if doc_url.startswith('http://'):
            if relative:
                raise ValueError('Relative links are only supported for local '
                                 'URLs (doc_url cannot start with "http://)"')
            searchindex_url = doc_url + '/' + searchindex
        else:
            searchindex_url = os.path.join(doc_url, searchindex)

        # detect if we are using relative links on a Windows system
        if os.name.lower() == 'nt' and not doc_url.startswith('http://'):
            if not relative:
                raise ValueError('You have to use relative=True for the local'
                                 'package on a Windows system.')
            self._is_windows = True
        else:
            self._is_windows = False

        # download and initialize the search index
        sindex = get_data(searchindex_url)
        filenames, objects = parse_sphinx_searchindex(sindex)

        self._searchindex = dict(filenames=filenames, objects=objects)

    def _get_link(self, cobj):
        """Get a valid link, False if not found"""

        fname_idx = None
        full_name = cobj['module_short'] + '.' + cobj['name']
        if full_name in self._searchindex['objects']:
            value = self._searchindex['objects'][full_name]
            if isinstance(value, dict):
                value = value[value.keys()[0]]
            fname_idx = value[0]
        elif cobj['module_short'] in self._searchindex['objects']:
            value = self._searchindex['objects'][cobj['module_short']]
            if cobj['name'] in value.keys():
                fname_idx = value[cobj['name']][0]

        if fname_idx is not None:
            fname = self._searchindex['filenames'][fname_idx] + '.html'

            if self._is_windows:
                fname = fname.replace('/', '\\')
                link = os.path.join(self.doc_url, fname)
            else:
                link = posixpath.join(self.doc_url, fname)

            if link in self._page_cache:
                html = self._page_cache[link]
            else:
                html = get_data(link)
                self._page_cache[link] = html

            # test if cobj appears in page
            comb_names = [cobj['module_short'] + '.' + cobj['name']]
            if self.extra_modules_test is not None:
                for mod in self.extra_modules_test:
                    comb_names.append(mod + '.' + cobj['name'])
            url = False
            for comb_name in comb_names:
                if html.find(comb_name) >= 0:
                    url = link + '#' + comb_name
            link = url
        else:
            link = False

        return link

    def resolve(self, cobj, this_url):
        """Resolve the link to the documentation, returns None if not found

        Parameters
        ----------
        cobj : dict
            Dict with information about the "code object" for which we are
            resolving a link.
            cobi['name'] : function or class name (str)
            cobj['module_short'] : shortened module name (str)
            cobj['module'] : module name (str)
        this_url: str
            URL of the current page. Needed to construct relative URLs
            (only used if relative=True in constructor).

        Returns
        -------
        link : str | None
            The link (URL) to the documentation.
        """
        full_name = cobj['module_short'] + '.' + cobj['name']
        link = self._link_cache.get(full_name, None)
        if link is None:
            # we don't have it cached
            link = self._get_link(cobj)
            # cache it for the future
            self._link_cache[full_name] = link

        if link is False or link is None:
            # failed to resolve
            return None

        if self.relative:
            link = os.path.relpath(link, start=this_url)
            if self._is_windows:
                # replace '\' with '/' so it on the web
                link = link.replace('\\', '/')

            # for some reason, the relative link goes one directory too high up
            link = link[3:]

        return link


###############################################################################
rst_template = """

.. _example_%(short_fname)s:

%(docstring)s

**Python source code:** :download:`%(fname)s <%(fname)s>`

.. literalinclude:: %(fname)s
    :lines: %(end_row)s-
    """

plot_rst_template = """

.. _example_%(short_fname)s:

%(docstring)s

%(image_list)s

%(stdout)s

**Python source code:** :download:`%(fname)s <%(fname)s>`

.. literalinclude:: %(fname)s
    :lines: %(end_row)s-

**Total running time of the example:** %(time_elapsed) .2f seconds
    """

# The following strings are used when we have several pictures: we use
# an html div tag that our CSS uses to turn the lists into horizontal
# lists.
HLIST_HEADER = """
.. rst-class:: horizontal

"""

HLIST_IMAGE_TEMPLATE = """
    *

      .. image:: images/%s
            :scale: 47
"""

SINGLE_IMAGE = """
.. image:: images/%s
    :align: center
"""


def extract_docstring(filename, ignore_heading=False):
    """ Extract a module-level docstring, if any
    """
    if sys.version_info >= (3, 0):
        lines = open(filename, encoding='utf-8').readlines()
    else:
        lines = open(filename).readlines()
    start_row = 0
    if lines[0].startswith('#!'):
        lines.pop(0)
        start_row = 1
    docstring = ''
    first_par = ''
    line_iterator = iter(lines)
    tokens = tokenize.generate_tokens(lambda: next(line_iterator))
    for tok_type, tok_content, _, (erow, _), _ in tokens:
        tok_type = token.tok_name[tok_type]
        if tok_type in ('NEWLINE', 'COMMENT', 'NL', 'INDENT', 'DEDENT'):
            continue
        elif tok_type == 'STRING':
            docstring = eval(tok_content)
            # If the docstring is formatted with several paragraphs, extract
            # the first one:
            paragraphs = '\n'.join(
                line.rstrip() for line
                in docstring.split('\n')).split('\n\n')
            if paragraphs:
                if ignore_heading:
                    if len(paragraphs) > 1:
                        first_par = re.sub('\n', ' ', paragraphs[1])
                        first_par = ((first_par[:95] + '...')
                                     if len(first_par) > 95 else first_par)
                    else:
                        raise ValueError("Docstring not found by gallery",
                                         "Please check your example's layout",
                                         " and make sure it's correct")
                else:
                    first_par = paragraphs[0]

        break
    return docstring, first_par, erow + 1 + start_row


def generate_example_rst(app):
    """ Generate the list of examples, as well as the contents of
        examples.
    """
    root_dir = os.path.join(app.builder.srcdir, 'auto_examples')
    example_dir = os.path.abspath(app.builder.srcdir + '/../../' + 'examples')
    try:
        plot_gallery = eval(app.builder.config.plot_gallery)
    except TypeError:
        plot_gallery = bool(app.builder.config.plot_gallery)
    if not os.path.exists(example_dir):
        os.makedirs(example_dir)
    if not os.path.exists(root_dir):
        os.makedirs(root_dir)

    # we create an index.rst with all examples
    fhindex = open(os.path.join(root_dir, 'index.rst'), 'w')
    fhindex.write("""\



.. raw:: html


    <style type="text/css">

    .figure {
        float: left;
        margin: 10px;
        -webkit-border-radius: 10px; /* Saf3-4, iOS 1-3.2, Android <1.6 */
        -moz-border-radius: 10px; /* FF1-3.6 */
        border-radius: 10px; /* Opera 10.5, IE9, Saf5, Chrome, FF4, iOS 4, Android 2.1+ */
        border: 2px solid #fff;
        background-color: white;
        /* --> Thumbnail image size */
        width: 150px;
        height: 100px;
        -webkit-background-size: 150px 100px; /* Saf3-4 */
        -moz-background-size: 150px 100px; /* FF3.6 */
    }

    .figure img {
        display: inline;
    }

    div.docstringWrapper p.caption {
        display: block;
        -webkit-box-shadow: 0px 0px 20px rgba(0, 0, 0, 0.0);
        -moz-box-shadow: 0px 0px 20px rgba(0, 0, 0, .0); /* FF3.5 - 3.6 */
        box-shadow: 0px 0px 20px rgba(0, 0, 0, 0.0); /* Opera 10.5, IE9, FF4+, Chrome 10+ */
        padding: 0px;
        border: white;
    }

    div.docstringWrapper p {
        display: none;
        background-color: white;
        -webkit-box-shadow: 0px 0px 20px rgba(0, 0, 0, 1.00);
        -moz-box-shadow: 0px 0px 20px rgba(0, 0, 0, 1.00); /* FF3.5 - 3.6 */
        box-shadow: 0px 0px 20px rgba(0, 0, 0, 1.00); /* Opera 10.5, IE9, FF4+, Chrome 10+ */
        padding: 13px;
        margin-top: 0px;
        border-style: solid;
        border-width: 1px;
    }


    </style>


.. raw:: html


        <script type="text/javascript">

        function animateClone(e){
          var position;
          position = $(this).position();
          var clone = $(this).closest('.thumbnailContainer').find('.clonedItem');
          var clone_fig = clone.find('.figure');
          clone.css("left", position.left - 70).css("top", position.top - 70).css("position", "absolute").css("z-index", 1000).css("background-color", "white");

          var cloneImg = clone_fig.find('img');

          clone.show();
          clone.animate({
                height: "270px",
                width: "320px"
            }, 0
          );
          cloneImg.css({
                'max-height': "200px",
                'max-width': "280px"
          });
          cloneImg.animate({
                height: "200px",
                width: "280px"
            }, 0
           );
          clone_fig.css({
               'margin-top': '20px',
          });
          clone_fig.show();
          clone.find('p').css("display", "block");
          clone_fig.css({
               height: "240",
               width: "305px"
          });
          cloneP_height = clone.find('p.caption').height();
          clone_fig.animate({
               height: (200 + cloneP_height)
           }, 0
          );

          clone.bind("mouseleave", function(e){
              clone.animate({
                  height: "100px",
                  width: "150px"
              }, 10, function(){$(this).hide();});
              clone_fig.animate({
                  height: "100px",
                  width: "150px"
              }, 10, function(){$(this).hide();});
          });
        } //end animateClone()


        $(window).load(function () {
            $(".figure").css("z-index", 1);

            $(".docstringWrapper").each(function(i, obj){
                var clone;
                var $obj = $(obj);
                clone = $obj.clone();
                clone.addClass("clonedItem");
                clone.appendTo($obj.closest(".thumbnailContainer"));
                clone.hide();
                $obj.bind("mouseenter", animateClone);
            }); // end each
        }); // end

        </script>



Examples
========

.. _examples-index:
""")
    # Here we don't use an os.walk, but we recurse only twice: flat is
    # better than nested.
    generate_dir_rst('.', fhindex, example_dir, root_dir, plot_gallery)
    for dir in sorted(os.listdir(example_dir)):
        if os.path.isdir(os.path.join(example_dir, dir)):
            generate_dir_rst(dir, fhindex, example_dir, root_dir, plot_gallery)
    fhindex.flush()


def extract_line_count(filename, target_dir):
    # Extract the line count of a file
    example_file = os.path.join(target_dir, filename)
    if sys.version_info >= (3, 0):
        lines = open(example_file, encoding='utf-8').readlines()
    else:
        lines = open(example_file).readlines()
    start_row = 0
    if lines and lines[0].startswith('#!'):
        lines.pop(0)
        start_row = 1
    line_iterator = iter(lines)
    tokens = tokenize.generate_tokens(lambda: next(line_iterator))
    check_docstring = True
    erow_docstring = 0
    for tok_type, _, _, (erow, _), _ in tokens:
        tok_type = token.tok_name[tok_type]
        if tok_type in ('NEWLINE', 'COMMENT', 'NL', 'INDENT', 'DEDENT'):
            continue
        elif ((tok_type == 'STRING') and check_docstring):
            erow_docstring = erow
            check_docstring = False
    return erow_docstring+1+start_row, erow+1+start_row


def line_count_sort(file_list, target_dir):
    # Sort the list of examples by line-count
    new_list = [x for x in file_list if x.endswith('.py')]
    unsorted = np.zeros(shape=(len(new_list), 2))
    unsorted = unsorted.astype(np.object)
    for count, exmpl in enumerate(new_list):
        docstr_lines, total_lines = extract_line_count(exmpl, target_dir)
        unsorted[count][1] = total_lines - docstr_lines
        unsorted[count][0] = exmpl
    index = np.lexsort((unsorted[:, 0].astype(np.str),
                        unsorted[:, 1].astype(np.float)))
    if not len(unsorted):
        return []
    return np.array(unsorted[index][:, 0]).tolist()


def generate_dir_rst(dir, fhindex, example_dir, root_dir, plot_gallery):
    """ Generate the rst file for an example directory.
    """
    if not dir == '.':
        target_dir = os.path.join(root_dir, dir)
        src_dir = os.path.join(example_dir, dir)
    else:
        target_dir = root_dir
        src_dir = example_dir
    if not os.path.exists(os.path.join(src_dir, 'README.txt')):
        print(80 * '_')
        print('Example directory %s does not have a README.txt file' %
              src_dir)
        print('Skipping this directory')
        print(80 * '_')
        return
    fhindex.write("""


%s


""" % open(os.path.join(src_dir, 'README.txt')).read())
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)
    sorted_listdir = line_count_sort(os.listdir(src_dir),
                                     src_dir)
    for fname in sorted_listdir:
        if fname.endswith('py'):
            generate_file_rst(fname, target_dir, src_dir, plot_gallery)
            new_fname = os.path.join(src_dir, fname)
            _, fdocstring, _ = extract_docstring(new_fname, True)
            thumb = os.path.join(dir, 'images', 'thumb', fname[:-3] + '.png')
            link_name = os.path.join(dir, fname).replace(os.path.sep, '_')
            fhindex.write("""

.. raw:: html


    <div class="thumbnailContainer">
        <div class="docstringWrapper">


""")

            fhindex.write('.. figure:: %s\n' % thumb)
            if link_name.startswith('._'):
                link_name = link_name[2:]
            if dir != '.':
                fhindex.write('   :target: ./%s/%s.html\n\n' % (dir,
                                                                fname[:-3]))
            else:
                fhindex.write('   :target: ./%s.html\n\n' % link_name[:-3])
            fhindex.write("""   :ref:`example_%s`


.. raw:: html


    <p>%s
    </p></div>
    </div>


.. toctree::
   :hidden:

   %s/%s

""" % (link_name, fdocstring, dir, fname[:-3]))
    fhindex.write("""
.. raw:: html

    <div style="clear: both"></div>
    """)  # clear at the end of the section

# modules for which we embed links into example code
DOCMODULES = ['pyart', 'matplotlib', 'numpy', 'scipy']


def make_thumbnail(in_fname, out_fname, width, height):
    """Make a thumbnail with the same aspect ratio centered in an
       image with a given width and height
    """
    img = Image.open(in_fname)
    width_in, height_in = img.size
    scale_w = width / float(width_in)
    scale_h = height / float(height_in)

    if height_in * scale_w <= height:
        scale = scale_w
    else:
        scale = scale_h

    width_sc = int(round(scale * width_in))
    height_sc = int(round(scale * height_in))

    # resize the image
    img.thumbnail((width_sc, height_sc), Image.ANTIALIAS)

    # insert centered
    thumb = Image.new('RGB', (width, height), (255, 255, 255))
    pos_insert = ((width - width_sc) // 2, (height - height_sc) // 2)
    thumb.paste(img, pos_insert)

    thumb.save(out_fname)


def get_short_module_name(module_name, obj_name):
    """ Get the shortest possible module name """
    parts = module_name.split('.')
    short_name = module_name
    for i in range(len(parts) - 1, 0, -1):
        short_name = '.'.join(parts[:i])
        try:
            exec('from %s import %s' % (short_name, obj_name))
        except ImportError:
            # get the last working module name
            short_name = '.'.join(parts[:(i + 1)])
            break
    return short_name


def generate_file_rst(fname, target_dir, src_dir, plot_gallery):
    """ Generate the rst file for a given example.
    """
    base_image_name = os.path.splitext(fname)[0]
    image_fname = '%s_%%s.png' % base_image_name

    this_template = rst_template
    last_dir = os.path.split(src_dir)[-1]
    # to avoid leading . in file names, and wrong names in links
    if last_dir == '.' or last_dir == 'examples':
        last_dir = ''
    else:
        last_dir += '_'
    short_fname = last_dir + fname
    src_file = os.path.join(src_dir, fname)
    example_file = os.path.join(target_dir, fname)
    shutil.copyfile(src_file, example_file)

    # The following is a list containing all the figure names
    figure_list = []

    image_dir = os.path.join(target_dir, 'images')
    thumb_dir = os.path.join(image_dir, 'thumb')
    if not os.path.exists(image_dir):
        os.makedirs(image_dir)
    if not os.path.exists(thumb_dir):
        os.makedirs(thumb_dir)
    image_path = os.path.join(image_dir, image_fname)
    stdout_path = os.path.join(image_dir,
                               'stdout_%s.txt' % base_image_name)
    time_path = os.path.join(image_dir,
                             'time_%s.txt' % base_image_name)
    thumb_file = os.path.join(thumb_dir, fname[:-3] + '.png')
    time_elapsed = 0
    if plot_gallery and fname.startswith('plot'):
        # generate the plot as png image if file name
        # starts with plot and if it is more recent than an
        # existing image.
        first_image_file = image_path % 1
        if os.path.exists(stdout_path):
            stdout = open(stdout_path).read()
        else:
            stdout = ''
        if os.path.exists(time_path):
            time_elapsed = float(open(time_path).read())

        if not os.path.exists(first_image_file) or \
           os.stat(first_image_file).st_mtime <= os.stat(src_file).st_mtime:
            # We need to execute the code
            print('plotting %s' % fname)
            t0 = time()
            import matplotlib.pyplot as plt
            plt.close('all')
            cwd = os.getcwd()
            try:
                # First CD in the original example dir, so that any file
                # created by the example get created in this directory
                orig_stdout = sys.stdout
                os.chdir(os.path.dirname(src_file))
                my_buffer = StringIO()
                my_stdout = Tee(sys.stdout, my_buffer)
                sys.stdout = my_stdout
                my_globals = {'pl': plt}
                execfile(os.path.basename(src_file), my_globals)
                time_elapsed = time() - t0
                sys.stdout = orig_stdout
                my_stdout = my_buffer.getvalue()

                # get variables so we can later add links to the documentation
                example_code_obj = {}
                for var_name, var in my_globals.items():
                    if not hasattr(var, '__module__'):
                        continue
                    if not isinstance(var.__module__, basestring):
                        continue
                    if var.__module__.split('.')[0] not in DOCMODULES:
                        continue

                    # get the type as a string with other things stripped
                    tstr = str(type(var))
                    tstr = (tstr[tstr.find('\'')
                            + 1:tstr.rfind('\'')].split('.')[-1])
                    # get shortened module name
                    module_short = get_short_module_name(var.__module__,
                                                         tstr)
                    cobj = {'name': tstr, 'module': var.__module__,
                            'module_short': module_short,
                            'obj_type': 'object'}
                    example_code_obj[var_name] = cobj

                # find functions so we can later add links to the documentation
                funregex = re.compile('[\w.]+\(')
                with open(src_file, 'rt') as fid:
                    for line in fid.readlines():
                        if line.startswith('#'):
                            continue
                        for match in funregex.findall(line):
                            fun_name = match[:-1]

                            try:
                                exec('this_fun = %s' % fun_name, my_globals)
                            except Exception as err:
                                # Here, we were not able to execute the
                                # previous statement, either because the
                                # fun_name was not a function but a statement
                                # (print), or because the regexp didn't
                                # catch the whole function name :
                                #    eg:
                                #       X = something().blah()
                                # will work for something, but not blah.

                                continue
                            this_fun = my_globals['this_fun']
                            if not callable(this_fun):
                                continue
                            if not hasattr(this_fun, '__module__'):
                                continue
                            if not isinstance(this_fun.__module__, basestring):
                                continue
                            if (this_fun.__module__.split('.')[0]
                                    not in DOCMODULES):
                                continue

                            # get shortened module name
                            fun_name_short = fun_name.split('.')[-1]
                            module_short = get_short_module_name(
                                this_fun.__module__, fun_name_short)
                            cobj = {'name': fun_name_short,
                                    'module': this_fun.__module__,
                                    'module_short': module_short,
                                    'obj_type': 'function'}
                            example_code_obj[fun_name] = cobj
                fid.close()

                if len(example_code_obj) > 0:
                    # save the dictionary, so we can later add hyperlinks
                    codeobj_fname = example_file[:-3] + '_codeobj.pickle'
                    with open(codeobj_fname, 'wb') as fid:
                        pickle.dump(example_code_obj, fid,
                                    pickle.HIGHEST_PROTOCOL)
                    fid.close()

                if '__doc__' in my_globals:
                    # The __doc__ is often printed in the example, we
                    # don't with to echo it
                    my_stdout = my_stdout.replace(
                        my_globals['__doc__'],
                        '')
                my_stdout = my_stdout.strip()
                if my_stdout:
                    stdout = '**Script output**::\n\n  %s\n\n' % (
                        '\n  '.join(my_stdout.split('\n')))
                open(stdout_path, 'w').write(stdout)
                open(time_path, 'w').write('%f' % time_elapsed)
                os.chdir(cwd)

                # In order to save every figure we have two solutions :
                # * iterate from 1 to infinity and call plt.fignum_exists(n)
                #   (this requires the figures to be numbered
                #    incrementally: 1, 2, 3 and not 1, 2, 5)
                # * iterate over [fig_mngr.num for fig_mngr in
                #   matplotlib._pylab_helpers.Gcf.get_all_fig_managers()]
                for fig_num in (fig_mngr.num for fig_mngr in
                    matplotlib._pylab_helpers.Gcf.get_all_fig_managers()):
                    # Set the fig_num figure as the current figure as we can't
                    # save a figure that's not the current figure.
                    plt.figure(fig_num)
                    plt.savefig(image_path % fig_num)
                    figure_list.append(image_fname % fig_num)
            except:
                print(80 * '_')
                print('%s is not compiling:' % fname)
                traceback.print_exc()
                print(80 * '_')
            finally:
                os.chdir(cwd)
                sys.stdout = orig_stdout

            print(" - time elapsed : %.2g sec" % time_elapsed)
        else:
            figure_list = [f[len(image_dir):]
                            for f in glob.glob(image_path % '[1-9]')]
                            #for f in glob.glob(image_path % '*')]

        # generate thumb file
        this_template = plot_rst_template
        if os.path.exists(first_image_file):
            make_thumbnail(first_image_file, thumb_file, 400, 280)

    if not os.path.exists(thumb_file):
        # create something to replace the thumbnail
        make_thumbnail('images/no_image.png', thumb_file, 200, 140)

    docstring, short_desc, end_row = extract_docstring(example_file)

    # Depending on whether we have one or more figures, we're using a
    # horizontal list or a single rst call to 'image'.
    if len(figure_list) == 1:
        figure_name = figure_list[0]
        image_list = SINGLE_IMAGE % figure_name.lstrip('/')
    else:
        image_list = HLIST_HEADER
        for figure_name in figure_list:
            image_list += HLIST_IMAGE_TEMPLATE % figure_name.lstrip('/')

    f = open(os.path.join(target_dir, fname[:-2] + 'rst'), 'w')
    f.write(this_template % locals())
    f.flush()


def embed_code_links(app, exception):
    """Embed hyperlinks to documentation into example code"""
    try:
        if exception is not None:
            return
        print('Embedding documentation hyperlinks in examples..')

        # Add resolvers for the packages for which we want to show links
        doc_resolvers = {}
        doc_resolvers['pyart'] = SphinxDocLinkResolver(app.builder.outdir,
                                                       relative=True)

        doc_resolvers['matplotlib'] = SphinxDocLinkResolver(
            'http://matplotlib.org')

        doc_resolvers['numpy'] = SphinxDocLinkResolver(
            'http://docs.scipy.org/doc/numpy-1.6.0')

        doc_resolvers['scipy'] = SphinxDocLinkResolver(
            'http://docs.scipy.org/doc/scipy-0.11.0/reference')

        example_dir = os.path.join(app.builder.srcdir, 'auto_examples')
        html_example_dir = os.path.abspath(os.path.join(app.builder.outdir,
                                                        'auto_examples'))

        # patterns for replacement
        link_pattern = '<a href="%s">%s</a>'
        orig_pattern = '<span class="n">%s</span>'
        period = '<span class="o">.</span>'

        for dirpath, _, filenames in os.walk(html_example_dir):
            for fname in filenames:
                print('\tprocessing: %s' % fname)
                full_fname = os.path.join(html_example_dir, dirpath, fname)
                subpath = dirpath[len(html_example_dir) + 1:]
                pickle_fname = os.path.join(example_dir, subpath,
                                            fname[:-5] + '_codeobj.pickle')

                if os.path.exists(pickle_fname):
                    # we have a pickle file with the objects to embed links for
                    with open(pickle_fname, 'rb') as fid:
                        example_code_obj = pickle.load(fid)
                    fid.close()
                    str_repl = {}
                    # generate replacement strings with the links
                    for name, cobj in example_code_obj.iteritems():
                        this_module = cobj['module'].split('.')[0]

                        if this_module not in doc_resolvers:
                            continue

                        link = doc_resolvers[this_module].resolve(cobj,
                                                                  full_fname)
                        if link is not None:
                            parts = name.split('.')
                            name_html = orig_pattern % parts[0]
                            for part in parts[1:]:
                                name_html += period + orig_pattern % part
                            str_repl[name_html] = link_pattern % (link, name_html)
                    # do the replacement in the html file
                    if len(str_repl) > 0:
                        with open(full_fname, 'rb') as fid:
                            lines_in = fid.readlines()
                        with open(full_fname, 'wb') as fid:
                            for line in lines_in:
                                line = line.decode('utf-8')
                                for name, link in str_repl.iteritems():
                                    line = line.replace(name, link)
                                fid.write(line.encode('utf-8'))
    except HTTPError as e:
        print("The following HTTP Error has occurred:\n")
        print(e.code)
    except URLError as e:
        print("\n...\n"
              "Warning: Embedding the documentation hyperlinks requires "
              "internet access.\nPlease check your network connection.\n"
              "Unable to continue embedding due to a URL Error: \n")
        print(e.args)
    print('[done]')


def setup(app):
    app.connect('builder-inited', generate_example_rst)
    app.add_config_value('plot_gallery', True, 'html')

    # embed links after build is finished
    # XXX disable documentation hyperlinks in examples for now
    #app.connect('build-finished', embed_code_links)

    # Sphinx hack: sphinx copies generated images to the build directory
    #  each time the docs are made.  If the desired image name already
    #  exists, it appends a digit to prevent overwrites.  The problem is,
    #  the directory is never cleared.  This means that each time you build
    #  the docs, the number of images in the directory grows.
    #
    # This question has been asked on the sphinx development list, but there
    #  was no response: http://osdir.com/ml/sphinx-dev/2011-02/msg00123.html
    #
    # The following is a hack that prevents this behavior by clearing the
    #  image build directory each time the docs are built.  If sphinx
    #  changes their layout between versions, this will not work (though
    #  it should probably not cause a crash).  Tested successfully
    #  on Sphinx 1.0.7
    build_image_dir = 'build/html/_images'
    if os.path.exists(build_image_dir):
        filelist = os.listdir(build_image_dir)
        for filename in filelist:
            if filename.endswith('png'):
                os.remove(os.path.join(build_image_dir, filename))



"""
==============================
Filters (:mod:`pyart.filters`)
==============================

.. currentmodule:: pyart.filters

Classes for specifying what gates are included and excluded from routines.

Filtering radar data
====================

.. autosummary::
    :toctree: generated/

    GateFilter
    moment_based_gate_filter
    moment_and_texture_based_gate_filter
    snr_based_gate_filter
    visibility_based_gate_filter
    class_based_gate_filter
    temp_based_gate_filter
    iso0_based_gate_filter
    birds_gate_filter

"""

from .gatefilter import GateFilter, moment_based_gate_filter
from .gatefilter import moment_and_texture_based_gate_filter
from .gatefilter import snr_based_gate_filter, visibility_based_gate_filter
from .gatefilter import temp_based_gate_filter, class_based_gate_filter
from .gatefilter import iso0_based_gate_filter, birds_gate_filter

__all__ = [s for s in dir() if not s.startswith('_')]



"""
Stream friendly, simple compression library, built around
iterators. See L{compress} and L{decompress} for the easiest way to
get started.

Basically implemented after the TIFF implementation of LZW, as described here:
U{http://www.fileformat.info/format/tiff/corion-lzw.htm}
But we included adjustments for more control codes and larger code sizes.

In an even-nuttier-shell, lzw compresses input bytes with integer
codes. Starting with codes 0-255 that code to themselves, and three
control codes, we work our way through a stream of bytes. When we
encounter a pair of codes c1,c2 we add another entry to our code table
with the lowest available code and the value value(c1) + value(c2)[0]

Of course, there are details :)

The Details
===========

    Our control codes are
        - END_OF_INFO_CODE (codepoint 256). This code is reserved for
          encoder/decoders over the integer codepoint stream (like the
          mechanical bit that unpacks bits into codepoints)
        - BUMB_CODE (codepoint 257). When this code is encountered we
          increase the bit size of the codepoints by 1.
        - CLEAR_CODE (codepoint 258). When this code is encountered, we flush
          the codebook and start over.

    When dealing with bytes, codes are emitted as variable
    length bit strings packed into the stream of bytes.

    codepoints are written with varying length
        - initially 9 bits
        - at 512   entries 10 bits
        - at 1024  entries 11 bits
        - at 2048  entries 12 bits
        - at 4096  entries 13 bits
        - at 8192  entries 14 bits
        - at 16384 entries 15 bits
        - with max 32768 entries in a table (including EOI, CLEAR and BUMP codes)

    code points are stored with their MSB in the most significant bit
    available in the output character.

>>> import lzw15
>>>
>>> mybytes = lzw15.readbytes("README.txt")
>>> lessbytes = lzw15.compress(mybytes)
>>> newbytes = b"".join(lzw15.decompress(lessbytes))
>>> oldbytes = b"".join(lzw15.readbytes("README.txt"))
>>> oldbytes == newbytes
True


"""

import struct
import operator
import itertools

END_OF_INFO_CODE = 256
BUMP_CODE = 257
NEXT_BUMP_CODE = 511
CLEAR_CODE = 258

DEFAULT_MIN_BITS = 9
DEFAULT_MAX_BITS = 15

MAX_CODE = 2**DEFAULT_MAX_BITS

FIRST_CODE = 259

def compress(plaintext_bytes):
    """
    Given an iterable of bytes, returns a (hopefully shorter) iterable
    of bytes that you can store in a file or pass over the network or
    what-have-you, and later use to get back your original bytes with
    L{decompress}. This is the best place to start using this module.
    """
    encoder = ByteEncoder()
    return encoder.encodetobytes(plaintext_bytes)


def decompress(compressed_bytes):
    """
    Given an iterable of bytes that were the result of a call to
    L{compress}, returns an iterator over the uncompressed bytes.
    """
    decoder = ByteDecoder()
    return decoder.decodefrombytes(compressed_bytes)


class ByteEncoder():
    """
    Takes a stream of uncompressed bytes and produces a stream of
    compressed bytes, usable by L{ByteDecoder}. Combines an L{Encoder}
    with a L{BitPacker}.


    >>> import lzw15
    >>>
    >>> enc = lzw15.ByteEncoder(12)
    >>> bigstr = b"gabba gabba yo gabba gabba gabba yo gabba gabba gabba yo gabba gabba gabba yo"
    >>> encoding = enc.encodetobytes(bigstr)
    >>> encoded = b"".join( b for b in encoding )
    >>> encoded == b'3\\x98LF#\\x08\\x82\\x05\\x04\\x83\\x1eM\\xf0x\\x1c
    >>>            \\x16\\x1b\\t\\x88C\\xe1q(4"\\x1f\\x17\\x85C#1X\\xec.\\x00'
    True
    >>>
    >>> dec = lzw15.ByteDecoder()
    >>> decoding = dec.decodefrombytes(encoded)
    >>> decoded = b"".join(decoding)
    >>> decoded == bigstr
    True

    """

    def __init__(self, max_width=DEFAULT_MAX_BITS):
        """
        max_width is the maximum width in bits we want to see in the
        output stream of codepoints.
        """
        self._encoder = Encoder(max_code_size=2**max_width)
        self._packer = BitPacker(initial_code_size=self._encoder.code_size())


    def encodetobytes(self, bytesource):
        """
        Returns an iterator of bytes, adjusting our packed width
        between minwidth and maxwidth when it detects an overflow is
        about to occur. Dual of L{ByteDecoder.decodefrombytes}.
        """
        codepoints = self._encoder.encode(bytesource)
        codebytes = self._packer.pack(codepoints)

        return codebytes


class ByteDecoder():
    """
    Decodes, combines bit-unpacking and interpreting a codepoint
    stream, suitable for use with bytes generated by
    L{ByteEncoder}.

    See L{ByteDecoder} for a usage example.
    """
    def __init__(self):
        """
        """

        self._decoder = Decoder()
        self._unpacker = BitUnpacker(initial_code_size=self._decoder.code_size())


    def decodefrombytes(self, bytesource):
        """
        Given an iterator over BitPacked, Encoded bytes, Returns an
        iterator over the uncompressed bytes. Dual of
        L{ByteEncoder.encodetobytes}. See L{ByteEncoder} for an
        example of use.
        """
        codepoints = self._unpacker.unpack(bytesource)
        clearbytes = self._decoder.decode(codepoints)

        return clearbytes


class BitPacker():
    """
    Translates a stream of lzw codepoints into a variable width packed
    stream of bytes, for use by L{BitUnpacker}.  One of a (potential)
    set of encoders for a stream of LZW codepoints, intended to behave
    as closely to the TIFF variable-width encoding as possible.

    The inbound stream of integer lzw codepoints are packed into
    variable width bit fields, starting at the smallest number of bits
    it can and then increasing the bit width as it anticipates the LZW
    code size growing to overflow.

    This class knows all kinds of intimate things about how it's
    upstream codepoint processors work; it knows the control codes
    CLEAR_CODE and END_OF_INFO_CODE, and (more intimately still), it
    makes assumptions about the rate of growth of it's consumer's
    codebook. This is ok, as long as the underlying encoder/decoders
    don't know any intimate details about their BitPackers/Unpackers
    """

    def __init__(self, initial_code_size):
        """
        Takes an initial code book size (that is, the count of known
        codes at the beginning of encoding, or after a clear)
        """
        self._initial_code_size = initial_code_size


    def pack(self, codepoints):
        """
        Given an iterator of integer codepoints, returns an iterator
        over bytes containing the codepoints packed into varying
        lengths, with bit width growing to accomodate an input code
        that it assumes will grow by one entry per codepoint seen.

        Widths will be reset to the given initial_code_size when the
        LZW CLEAR_CODE or END_OF_INFO_CODE code appears in the input,
        and bytes following END_OF_INFO_CODE will be aligned to the
        next byte boundary.

        >>> import lzw, six
        >>> pkr = lzw.BitPacker(258)
        >>> [ b for b in pkr.pack([ 1, 257]) ] == [ struct.Struct(">B").pack(0),
        >>>  struct.Struct(">B").pack(0xC0), struct.Struct(">B").pack(0x40) ]
        True
        """

        tailbits = []
        codesize = self._initial_code_size

        minwidth = DEFAULT_MIN_BITS

        nextwidth = minwidth

        for pt in codepoints:

            newbits = inttobits(pt, nextwidth)

            tailbits = tailbits + newbits #Last bits in the list.

            # PAY ATTENTION. This calculation should be driven by the
            # size of the upstream codebook, right now we're just trusting
            # that everybody intends to follow the TIFF spec.

            codesize = codesize + 1

            if pt == END_OF_INFO_CODE:
                while len(tailbits) % 8:
                    tailbits.append(0)

            if pt == BUMP_CODE:
                nextwidth = nextwidth + 1

            elif codesize >= MAX_CODE:
                print("Attention: codesize >= MAX_CODE. Check the python lzw code!")

            while len(tailbits) > 8:
                nextbits = tailbits[:8]
                nextbytes = bitstobytes(nextbits)

                for bt in nextbytes:
                    yield struct.pack("B", bt)

                tailbits = tailbits[8:]

        if tailbits:
            tail = bitstobytes(tailbits)
            #print("TAIL:", tailbits)
            for bt in tail:
                yield struct.pack("B", bt)


class BitUnpacker():
    """
    An adaptive-width bit unpacker, intended to decode streams written
    by L{BitPacker} into integer codepoints. Like L{BitPacker}, knows
    about code size changes and control codes.
    """

    def __init__(self, initial_code_size):
        """
        initial_code_size is the starting size of the codebook
        associated with the to-be-unpacked stream.
        """
        self._initial_code_size = initial_code_size


    def unpack(self, bytesource):
        """
        Given an iterator of bytes, returns an iterator of integer
        code points. Auto-magically adjusts point width when it sees
        an almost-overflow in the input stream, or an LZW CLEAR_CODE
        or END_OF_INFO_CODE

        Trailing bits at the end of the given iterator, after the last
        codepoint, will be dropped on the floor.

        At the end of the iteration, or when an END_OF_INFO_CODE seen
        the unpacker will ignore the bits after the code until it
        reaches the next aligned byte. END_OF_INFO_CODE will *not*
        stop the generator, just reset the alignment and the width

        >>> import lzw, six
        >>> unpk = lzw.BitUnpacker(initial_code_size=258)
        >>> [ i for i in unpk.unpack([ struct.Struct(">B").pack(0),
        >>>  struct.Struct(">B").pack(0xC0), struct.Struct(">B").pack(0x40) ]) ]
        [1, 257]
        """
        bits = []
        offset = 0
        ignore = 0

        codesize = self._initial_code_size

        #Uncommented
        #minwidth = 8
        #while (1 << minwidth) < codesize:
        #    minwidth = minwidth + 1

        minwidth = DEFAULT_MIN_BITS
        pointwidth = minwidth

        for nextbit in bytestobits(bytesource):

            offset = (offset + 1) % 8

            if ignore > 0:
                ignore = ignore - 1
                continue

            bits.append(nextbit)

            if len(bits) == pointwidth:
                codepoint = intfrombits(bits)
                bits = []
                yield codepoint
                codesize = codesize + 1

                if codepoint == BUMP_CODE:
                    pointwidth = pointwidth + 1

                if codepoint == END_OF_INFO_CODE:
                    ignore = (8 - offset) % 8


class Decoder():
    """
    Uncompresses a stream of lzw code points, as created by
    L{Encoder}. Given a list of integer code points, with all
    unpacking foolishness complete, turns that list of codepoints into
    a list of uncompressed bytes. See L{BitUnpacker} for what this
    doesn't do.
    """
    def __init__(self):
        """
        Creates a new Decoder. Decoders should not be reused for
        different streams.
        """
        self._clear_codes()
        self.remainder = []
        self._count = 0

    def code_size(self):
        """
        Returns the current size of the Decoder's code book, that is,
        it's mapping of codepoints to byte strings. The return value of
        this method will change as the decode encounters more encoded
        input, or control codes.
        """
        return len(self._codepoints)


    def decode(self, codepoints):
        """
        Given an iterable of integer codepoints, yields the
        corresponding bytes, one at a time, as byte strings of length
        E{1}. Retains the state of the codebook from call to call, so
        if you have another stream, you'll likely need another
        decoder!

        Decoders will NOT handle END_OF_INFO_CODE (rather, they will
        handle the code by throwing an exception); END_OF_INFO should
        be handled by the upstream codepoint generator (see
        L{BitUnpacker}, for example)

        >>> import lzw
        >>> dec = lzw.Decoder()
        >>> result = b''.join(dec.decode([103, 97, 98, 98, 97, 32, 258, 260,
        >>>          262, 121, 111, 263, 259, 261, 256]))
        >>> result == b'gabba gabba yo gabba'
        True

        """
        codepoints = [cp for cp in codepoints]

        for cp in codepoints:
            if cp == BUMP_CODE:
                #print("BUMB_CODE reached")
                continue
            if cp == END_OF_INFO_CODE:
                #print("END_OF_INFO_CODE reached")
                return

            decoded = self._decode_codepoint(cp)

            for character in iter(decoded):
                # TODO optimize, casting back to bytes when bytes above
                yield struct.Struct(">B").pack(character)


    def _decode_codepoint(self, codepoint):
        """
        Will raise a ValueError if given an END_OF_INFORMATION
        code. EOI codes should be handled by callers if they're
        present in our source stream.

        >>> import lzw
        >>> dec = lzw.Decoder()
        >>> beforesize = dec.code_size()
        >>> dec._decode_codepoint(0x80) == b'\\x80'
        True
        >>> dec._decode_codepoint(0x81) == b'\\x81'
        True
        >>> beforesize + 1 == dec.code_size()
        True
        >>> dec._decode_codepoint(256) == b''
        True
        >>> beforesize == dec.code_size()
        True
        """
        ret = b''

        if codepoint == CLEAR_CODE:
            self._clear_codes()
        elif codepoint == END_OF_INFO_CODE: #END_OF_INFO_CODE is handled in decode routine
            raise ValueError("End of information code not supported " +
                             "directly by this Decoder")
        else:
            if codepoint in self._codepoints:
                ret = self._codepoints[codepoint]
                if self._prefix is not None:
                    self._codepoints[len(self._codepoints)] = (
                        self._prefix + struct.Struct(">B").pack(operator.getitem(ret, 0)))

            else:
                ret = self._prefix + struct.Struct(">B").pack(
                    operator.getitem(self._prefix, 0))
                self._codepoints[len(self._codepoints)] = ret

            self._prefix = ret

        return ret


    def _clear_codes(self):
        self._codepoints = dict(
            (pt, struct.pack("B", pt)) for pt in range(256))

        self._codepoints[END_OF_INFO_CODE] = END_OF_INFO_CODE
        self._codepoints[BUMP_CODE] = BUMP_CODE
        self._codepoints[CLEAR_CODE] = CLEAR_CODE

        self._prefix = None


class Encoder():
    """
    Given an iterator of bytes, returns an iterator of integer
    codepoints, suitable for use by L{Decoder}. The core of the
    "compression" side of lzw compression/decompression.
    """
    def __init__(self, max_code_size=MAX_CODE):
        """
        When the encoding codebook grows larger than max_code_size,
        the Encoder will clear its codebook and emit a CLEAR_CODE
        """

        self.closed = False

        self._max_code_size = max_code_size
        self._buffer = b''
        self._clear_codes()

        if max_code_size < self.code_size():
            raise ValueError(
                "Max code size too small, (must be at least {0})".format(
                    self.code_size()))


    def code_size(self):
        """
        Returns a count of the known codes, including codes that are
        implicit in the data but have not yet been produced by the
        iterator.
        """
        return len(self._prefixes)


    def flush(self):
        """
        Yields any buffered codepoints, followed by a CLEAR_CODE, and
        clears the codebook as a side effect.
        """

        # flushed = []

        if self._buffer:
            yield self._prefixes[self._buffer]
            self._buffer = b''

        yield END_OF_INFO_CODE
        self._clear_codes()


    def bump(self):
        """
        Yields the BUMP_CODE and increases the next_bumb_code value
        """
        yield BUMP_CODE

        self._current_code_bits = self._current_code_bits + 1
        self._next_bumb_code <<= 1
        self._next_bumb_code |= 1

        #print("Next_bump_code:", self._next_bumb_code)


    def encode(self, bytesource):
        """
        Given an iterator over bytes, yields the
        corresponding stream of codepoints.
        Will clear the codes at the end of the stream.

        >>> import lzw
        >>> enc = lzw.Encoder()
        >>> [ cp for cp in enc.encode(b"gabba gabba yo gabba") ]
        [103, 97, 98, 98, 97, 32, 258, 260, 262, 121, 111, 263, 259, 261, 256]

        """
        for b in bytesource:
            for point in self._encode_byte(b):
                yield point
            if self.code_size() > self._max_code_size:
                for pt in self.flush():
                    yield pt
            if self.code_size() > self._next_bumb_code:
                for pt in self.bump():
                    yield pt

        #Important to get last bytes correct
        for point in self.flush():
            yield point

    def _encode_byte(self, point):
        # Yields one or zero bytes, AND changes the internal state of
        # the codebook and prefix buffer.
        #
        # Unless you're in self.encode(), you almost certainly don't
        # want to call this.

        # In python3 iterating over the bytestring will return in codepoints,
        # we use the byte([]) constructor to conver this back into bytestring
        # so we can add to new_prefix and key the _prefixes by the bytestring.

        byte = point if isinstance(point, bytes) else struct.Struct(">B").pack(
            point)
        #print(byte)
        new_prefix = self._buffer
        #print(new_prefix)
        if new_prefix + byte in self._prefixes:
            new_prefix = new_prefix + byte
        elif new_prefix:
            encoded = self._prefixes[new_prefix]
            #print(encoded)
            #print(new_prefix + byte)
            self._add_code(new_prefix + byte)
            new_prefix = byte
            #print(self._prefixes)
            yield encoded

        self._buffer = new_prefix

    def _clear_codes(self):

        # Teensy hack, CLEAR_CODE and END_OF_INFO_CODE aren't
        # equal to any possible string.

        self._prefixes = dict(
            (struct.pack("B", codept), codept) for codept in range(256))
        self._prefixes[END_OF_INFO_CODE] = END_OF_INFO_CODE
        self._prefixes[BUMP_CODE] = BUMP_CODE
        self._prefixes[CLEAR_CODE] = CLEAR_CODE
        self._next_bumb_code = NEXT_BUMP_CODE
        self._current_code_bits = 9


    def _add_code(self, newstring):
        #print(len(self._prefixes))
        self._prefixes[newstring] = len(self._prefixes)
        #print(self._prefixes)


#########################################
# Conveniences.


def unpackbyte(b):
    """
    Given a one-byte long byte string, returns an integer. Equivalent
    to struct.unpack("B", b)
    """
    if isinstance(b, bytes):
        return operator.itemgetter(0)(b)
    return b


def filebytes(fileobj, buffersize=1024):
    """
    Convenience for iterating over the bytes in a file. Given a
    file-like object (with a read(int) method), returns an iterator
    over the bytes of that file.
    """
    buff = fileobj.read(buffersize)
    while buff:
        for byte in buff:
            yield byte
        buff = fileobj.read(buffersize)


def readbytes(filename, buffersize=1024):
    """
    Opens a file named by filename and iterates over the L{filebytes}
    found therein.  Will close the file when the bytes run out.
    """
    with open(filename, "rb") as infile:
        for byte in iter(filebytes(infile, buffersize)):
            yield struct.Struct(">B").pack(byte)  # TODO optimize, we are re-casting to bytes

def readbytes_fh(infile, buffersize=1024):
    """
    Iterates over the L{filebytes} of a file handler
    found therein.  Will close the file when the bytes run out.
    """
    for byte in iter(filebytes(infile, buffersize)):
        yield struct.Struct(">B").pack(byte)  # TODO optimize, we are re-casting to bytes


def writebytes(filename, bytesource):
    """
    Convenience for emitting the bytes we generate to a file. Given a
    filename, opens and truncates the file, dumps the bytes
    from bytesource into it, and closes it
    """

    with open(filename, "wb") as outfile:
        for bt in bytesource:
            outfile.write(bt)


def inttobits(anint, width=None):
    """
    Produces an array of booleans representing the given argument as
    an unsigned integer, MSB first. If width is given, will pad the
    MSBs to the given width (but will NOT truncate overflowing
    results)

    >>> import lzw15
    >>> lzw15.inttobits(304, width=16)
    [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0]

    """
    remains = anint
    retreverse = []
    while remains:
        retreverse.append(remains & 1)
        remains = remains >> 1

    retreverse.reverse()

    ret = retreverse
    if width is not None:
        ret_head = [0] * (width - len(ret))
        ret = ret_head + ret

    return ret


def intfrombits(bits):
    """
    Given a list of boolean values, interprets them as a binary
    encoded, MSB-first unsigned integer (with True == 1 and False
    == 0) and returns the result.

    >>> import lzw15
    >>> lzw15.intfrombits([ 1, 0, 0, 1, 1, 0, 0, 0, 0 ])
    304
    """
    ret = 0
    lsb_first = [b for b in bits]
    lsb_first.reverse()

    for bit_index, lsb_first_el in enumerate(lsb_first):
        if lsb_first_el:
            ret = ret | (1 << bit_index)

    return ret


def bytestobits(bytesource):
    """
    Breaks a given iterable of bytes into an iterable of boolean
    values representing those bytes as unsigned integers.

    >>> import lzw15
    >>> [ x for x in lzw15.bytestobits(b"\\x01\\x30") ]
    [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0]
    """
    for b in bytesource:

        value = unpackbyte(b)

        for bitplusone in range(8, 0, -1):
            bitindex = bitplusone - 1
            nextbit = 1 & (value >> bitindex)
            yield nextbit


def bitstobytes(bits):
    """
    Interprets an indexable list of booleans as bits, MSB first, to be
    packed into a list of integers from 0 to 256, MSB first, with LSBs
    zero-padded. Note this padding behavior means that round-trips of
    bytestobits(bitstobytes(x, width=W)) may not yield what you expect
    them to if W % 8 != 0

    Does *NOT* pack the returned values into a bytearray or the like.

    >>> import lzw15
    >>> bitstobytes([0, 0, 0, 0, 0, 0, 0, 0, "Yes, I'm True"]) == [ 0x00, 0x80 ]
    True
    >>> bitstobytes([0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 0]) == [ 0x01, 0x30 ]
    True
    """
    ret = []
    nextbyte = 0
    nextbit = 7
    for bit in bits:
        if bit:
            nextbyte = nextbyte | (1 << nextbit)

        if nextbit:
            nextbit = nextbit - 1
        else:
            ret.append(nextbyte)
            nextbit = 7
            nextbyte = 0

    if nextbit < 7:
        ret.append(nextbyte)
    return ret



#! /usr/bin/env python
"""
Make a small MDV file containing a grid with X and Y axes in degrees.

Partial grid is taken from full sized file 000000.mdv which is contained in
the 200202.MASTER15.mdv.tar.gz file available online at:
http://www2.mmm.ucar.edu/imagearchive/WSI/mdv/
"""
# The MDV RLE decoding routine ends at the end of the data stream even if not
# all data point have been read.  Therefore a file truncated at a break in the
# RLE can still be read but data past the last data point will be filled with
# random data.  Here we end after the first non-keyed RLE byte.
infile = open('000000.mdv', 'rb')
outfile = open('example_mdv_grid.mdv', 'wb')
outfile.write(infile.read(8134))
infile.close()
outfile.close()



"""
========================================
Radar Corrections (:mod:`pyart.correct`)
========================================

.. currentmodule:: pyart.correct

Correct radar fields.

Velocity unfolding
==================

.. autosummary::
    :toctree: generated/

    dealias_fourdd
    dealias_unwrap_phase
    dealias_region_based

Other corrections
=================

.. autosummary::
    :toctree: generated/

    calculate_attenuation
    calculate_attenuation_zphi
    calculate_attenuation_philinear
    phase_proc_lp
    det_sys_phase_ray
    correct_sys_phase
    smooth_phidp_single_window
    smooth_phidp_double_window
    despeckle_field
    correct_noise_rhohv
    correct_bias
    correct_visibility
    est_rhohv_rain
    est_zdr_precip
    est_zdr_snow
    selfconsistency_bias
    selfconsistency_bias2
    selfconsistency_kdp_phidp
    get_sun_hits
    get_sun_hits_psr
    get_sun_hits_ivic
    sun_retrieval
    phase_proc_lp_gf

Helper functions
================

.. autosummary::
    :toctree: generated/

    find_objects
    get_mask_fzl
    sun_power
    ptoa_to_sf
    solar_flux_lookup
    scanning_losses
    smooth_masked
    sun_position_pysolar
    sun_position_mfr

"""


from .dealias import dealias_fourdd
from .attenuation import calculate_attenuation, calculate_attenuation_zphi
from .attenuation import calculate_attenuation_philinear, get_mask_fzl
from .phase_proc import det_sys_phase_ray, correct_sys_phase
from .phase_proc import smooth_phidp_single_window, smooth_phidp_double_window
from .phase_proc import smooth_masked
from .phase_proc import phase_proc_lp, phase_proc_lp_gf

# for backwards compatibility GateFilter available in the correct namespace
from ..filters.gatefilter import GateFilter, moment_based_gate_filter
from .unwrap import dealias_unwrap_phase
from .region_dealias import dealias_region_based
from .despeckle import find_objects, despeckle_field
from .bias_and_noise import correct_noise_rhohv, correct_bias
from .bias_and_noise import correct_visibility
from .bias_and_noise import est_rhohv_rain, est_zdr_precip, est_zdr_snow
from .bias_and_noise import get_sun_hits, get_sun_hits_ivic, sun_retrieval
from .bias_and_noise import selfconsistency_bias, selfconsistency_bias2
from .bias_and_noise import selfconsistency_kdp_phidp, get_sun_hits_psr
from .sunlib import sun_power, solar_flux_lookup, ptoa_to_sf, scanning_losses
from .sunlib import sun_position_pysolar, sun_position_mfr
from .sunlib import gauss_fit, retrieval_result

__all__ = [s for s in dir() if not s.startswith('_')]



"""
pyart.io.odim_h5_writer.py
=========================

Utilities for writing ODIM hdf5 files.

.. autosummary::
    :toctree: generated/

    write_odim_h5
    _to_str
    _get_sec_since_epoch
    _tree
    _check_file_exists
    _create_odim_h5_file
    _create_odim_h5_grp
    _create_odim_h5_sub_grp
    _create_odim_h5_attr
    _create_odim_h5_dataset
    _map_radar_quantity
    _get_data_from_field
    _map_radar_to_how_dict

"""

import datetime
import calendar
import time
from warnings import warn
from collections import defaultdict
from pathlib import Path
import numpy as np

from ..exceptions import MissingOptionalDependency
from ..config import get_fillvalue

try:
    import h5py
    _H5PY_AVAILABLE = True
except ImportError:
    _H5PY_AVAILABLE = False

#from ..config import FileMetadata
#from ..core.radar import Radar
#from ..lazydict import LazyLoadDict


def write_odim_h5(filename, radar, field_names=None, physical=True,
                  compression="gzip", compression_opts=6):
    """
    Write a Radar object to a EUMETNET OPERA compliant HDF5 file.

    The files produced by this routine follow the EUMETNET OPERA information
    model:
    http://eumetnet.eu/wp-content/uploads/2017/01/OPERA_hdf_description_2014.pdf

    Supported features:
      - Writing PPIs: PVOL and SCAN objects
        - Different sweeps are saved in different dataset groups
      - Writing sectorized PPIs and SCANs: AZIM objects
      - Writing RHIs: ELEV objects

    Not yet supported:
      - Mixed datasets (how group always on top level)
      - Single ray data (e.g. from fixed staring mode)
      - Profiles


    Parameters
    ----------
    filename : str
        Filename of file to create.
    radar : Radar
        Radar object to process.
    field_names : list of str
        The list of fields from the radar object to save. If none all fields
        in the radar object will be saved.
    physical : Bool
        If true the physical values are stored. nodata parameter is equal to
        the _FillValue parameter in the field metadata or the default Py-ART
        fill value. If false the data is converted into binary values using a
        linear conversion. The gain and offset are either specified in the
        metadata of the field with keywords 'scale_factor' and 'add_offset' or
        calculated on the fly. keyword '_Write_as_dtype' specifies the
        datatype. It can be either 'uint8' or 'uint16'. The default datatype
        is uint8. The 'undetect' parameter is not used
    compression : str
        The type of compression for the datasets. Typical are "gzip" and "lzf".
    compression_opts : any
        The compression options. In the case of gzip is the level between 0 to
        9 (recomended 1 to 6). In the case of lzf there are not options.

    """

    #Initialize hdf5 file
    hdf_id = _create_odim_h5_file(filename)

    #Determine odim object, number of datasets (PPIs, RHIs) and
    # check if sweep mode does not change between different sweeps
    if radar.scan_type == 'rhi':
        odim_object = 'ELEV'
        n_datasets = np.size(radar.fixed_angle['data'])
    else:
        if radar.scan_type == 'ppi':
            n_datasets = radar.nsweeps
            if len(set(radar.fixed_angle['data'])) <= 1:
                odim_object = 'SCAN'
            else:
                if radar.sweep_mode is not None:
                    if len(set(radar.sweep_mode['data'])) <= 1:
                        nray_in_sweep = radar.rays_per_sweep['data'][0]
                        ray_angle_res = radar.ray_angle_res['data'][0]
                        if nray_in_sweep * ray_angle_res < 360:
                            odim_object = 'AZIM'
                        else:
                            odim_object = 'PVOL'
                    else:
                        warn("Radar sweep mode changes, which is not " +
                             "yet supported to write ODIM HDF5 file.")
                        return
                else:
                    warn("ODIM object could not be identified.")
                    return

    #Determine number of different data types per dataset and list of fields
    radar_field_names = list(radar.fields.keys())
    if field_names is not None:
        # check that all desired fields are in radar object
        aux_field_names = []
        for field_name in field_names:
            if field_name not in radar_field_names:
                warn(field_name+' not in radar object')
            else:
                aux_field_names.append(field_name)
        if np.size(aux_field_names) == 0:
            warn('No matching field names available')
            return
        field_names = aux_field_names

    else:
        field_names = list(radar.fields.keys())
    n_datatypes = len(field_names)

    #Create level 1 group structure
    where1_grp = _create_odim_h5_grp(hdf_id, '/where')
    what1_grp = _create_odim_h5_grp(hdf_id, '/what')
    how1_grp = _create_odim_h5_grp(hdf_id, '/how')

    dataset_grps = list()
    for i in range(n_datasets):
        name = 'dataset'+str(i+1)
        grp_id = _create_odim_h5_grp(hdf_id, name)
        dataset_grps.append(grp_id)

    #Write ODIM Conventions attribute
    _create_odim_h5_attr(hdf_id, 'Conventions', 'ODIM_H5/V2_2')

    #where - lon, lat, height
    lon = radar.longitude['data']
    lat = radar.latitude['data']
    height = radar.altitude['data']
    _create_odim_h5_attr(where1_grp, 'lon', lon)
    _create_odim_h5_attr(where1_grp, 'lat', lat)
    _create_odim_h5_attr(where1_grp, 'height', height)

    #what - version, date, time, source, object
    odim_version = _to_str(radar.metadata['version'])
    odim_source = _to_str(radar.metadata['source'])

    #Time
    odim_time_struct = time.strptime(
        radar.time['units'], "seconds since %Y-%m-%dT%H:%M:%SZ")
    odim_datetime_struct = datetime.datetime.utcfromtimestamp(
        calendar.timegm(odim_time_struct))

    #Time relative to center of first gate?
    odim_dt = radar.time['data'][0]
    odim_datetime_start = odim_datetime_struct + datetime.timedelta(
        seconds=odim_dt)

    odim_time = datetime.datetime.strftime(odim_datetime_start, "%H%M%S")
    odim_date = datetime.datetime.strftime(odim_datetime_start, "%Y%m%d")

    #Create and fill what1 group attributes
    _create_odim_h5_attr(what1_grp, 'time', odim_time)
    _create_odim_h5_attr(what1_grp, 'date', odim_date)
    _create_odim_h5_attr(what1_grp, 'version', odim_version)
    _create_odim_h5_attr(what1_grp, 'source', odim_source)
    _create_odim_h5_attr(what1_grp, 'object', odim_object)

    #How variables
    #General
    how_var_general = ['system', 'software', 'sw_verison']

    #Individual radar
    how_var_instrument = [
        'beamwidth', 'wavelength', 'rpm', 'elevspeed', 'pulsewidth',
        'RXbandwidth', 'lowprf', 'midprf', 'highprf', 'TXlossH', 'TXlossV'
        'injectlossH', 'injectlossV', 'RXlossH', 'RXlossV',
        'radomelossH', 'radomelossV', 'antgainH', 'antgainV', 'beamwH',
        'beamwV', 'gasattn', 'radconstH', 'radconstV', 'nomTXpower', 'TXpower'
        'powerdiff', 'phasediff', 'NI', 'Vsamples']

    # Map radar.metadata to how1_dict if entries are available
    if any(x in how_var_general for x in radar.metadata):
        if 'system' in radar.metadata:
            how_var_gen = ['system']
        if 'software' in radar.metadata:
            how_var_gen.append('software')
        if 'sw_verison' in radar.metadata:
            how_var_gen.append('sw_version')
        how1_gen_dict = _map_radar_to_how_dict(radar.metadata)
        for name in how_var_gen:
            _create_odim_h5_attr(how1_grp, name, how1_gen_dict[name])

    #Map radar.instrument_parameters to how1_dict
    if radar.instrument_parameters is not None:
        radar_ins_obj = radar.instrument_parameters
        how1_ins_dict = _map_radar_to_how_dict(radar_ins_obj)
    else:
        how1_ins_dict = None
        warn("Instrument parameters not available in radar object")

    #Map radar.radar_calibration to how1_dict
    if radar.radar_calibration is not None:
        radar_cal_obj = radar.radar_calibration
        how1_cal_dict = _map_radar_to_how_dict(radar_cal_obj)
    else:
        how1_cal_dict = None
        warn("Radar calibration parameters not available in radar object")

    for name in how_var_instrument:
        if how1_ins_dict is not None:
            if name in how1_ins_dict:
                _create_odim_h5_attr(how1_grp, name, how1_ins_dict[name])
        elif how1_cal_dict is not None:
            if name in how1_cal_dict:
                _create_odim_h5_attr(how1_grp, name, how1_cal_dict[name])

    #Create level 2 group structure
    datatype_grps = [] # 2D list datatype group ids
    where2_grps = list() # 1D list of where groups
    what2_grps = list() # 1D list of what groups
    how2_grps = list() # 1D list of how groups

    for i in range(n_datasets):
        where2_id = _create_odim_h5_sub_grp(dataset_grps[i], 'where')
        where2_grps.append(where2_id)
        what2_id = _create_odim_h5_sub_grp(dataset_grps[i], 'what')
        what2_grps.append(what2_id)
        how2_id = _create_odim_h5_sub_grp(dataset_grps[i], 'how')
        how2_grps.append(how2_id)

        datatype_grps.append([]) #empty list for each row

        for j in range(n_datatypes):
            name = 'data'+str(j+1)
            datatype_ind = _create_odim_h5_sub_grp(dataset_grps[i], name)
            datatype_grps[i].append(datatype_ind)

    # Dataset specific what header Attributes
    what_var_dataset = ['product', 'prodpar', 'startdate', 'starttime',
                        'enddate', 'endtime']

    # Data attributes
    what_var_data = ['quantity', 'gain', 'offset', 'nodata', 'undetect']

    # Supported scan types
    odim_product_list = ['PPI', 'RHI']

    if odim_object in ['PVOL', 'AZIM', 'SCAN', 'ELEV']:
        #where2
        where2_variables = ['nbins', 'rstart', 'rscale', 'nrays']
        where2_dict = {}

        where2_dict['nbins'] = np.int64(np.repeat(radar.ngates, n_datasets))
        where2_dict['rstart'] = np.repeat(
            np.double((radar.range['data'][0])/(1000.)), n_datasets)  #[km]
        where2_dict['nrays'] = np.int64(radar.rays_per_sweep['data'])

        if len(set(radar.range['data'][1:] - radar.range['data'][0:-1])) <= 1:
            range_resolution = np.double(
                radar.range['data'][1]-radar.range['data'][0])
        else:
            range_resolution = np.median(
                radar.range['data'][1:] - radar.range['data'][0:-1])
            warn("Radar range resolution changes between gates. " +
                 "The median resolution is taken as reference.")
        where2_dict['rscale'] = np.repeat(range_resolution, n_datasets)

        if odim_object in ['PVOL', 'AZIM', 'SCAN']:
            where2_variables.extend(['elangle', 'a1gate'])
            where2_dict['elangle'] = np.double(radar.fixed_angle['data'])
            where2_dict['a1gate'] = np.int64(
                radar.sweep_start_ray_index['data'])

        #Sector Specific
        elif odim_object == 'AZIM':
            where2_variables.extend(['startaz', 'stopaz'])
            start_az = []
            stop_az = []
            for i in radar.sweep_start_ray_index['data']:
                start_az_tmp = radar.azimuth['data'][i]
                start_az.append(start_az_tmp)
            for i in radar.sweep_end_ray_index['data']:
                stop_az_tmp = radar.azimuth['data'][i]
                stop_az.append(stop_az_tmp)
            where2_dict['startaz'] = np.double(start_az)
            where2_dict['stopaz'] = np.double(stop_az)

        #RHI specific
        elif odim_object == 'ELEV':
            where2_variables.extend(['az_angle', 'angles', 'range'])
            where2_dict['range'] = np.repeat(
                np.double((np.max(radar.range['data']) +
                           (radar.range['data'][0])) / 1000.),
                n_datasets) #[km] to end of last gate

            az_angle = []
            for i in range(n_datasets):
                az_angle_tmp = radar.fixed_angle['data'][i]
                az_angle.append(az_angle_tmp)
            where2_dict['az_angle'] = np.double(az_angle)
            angles_mx = [
                [0 for x in range(radar.nrays)] for y in range(n_datasets)]
            for i in range(n_datasets):
                angles_mx[i][:] = radar.elevation['data']
            where2_dict['angles'] = np.double(angles_mx)

        # Write where2 group attributes
        ind = 0
        for i in where2_grps:
            for name in where2_variables:
                if where2_dict[name] is not None:
                    _create_odim_h5_attr(i, name, where2_dict[name][ind])
            ind = ind+1

        # what2
        scan_type = radar.scan_type
        what2_dict = {}
        what2_dict['product'] = np.repeat(scan_type.upper(), n_datasets)
        if scan_type.upper() in odim_product_list:
            if scan_type.upper() == 'PPI':
                # Eleveation angle/s (deg)
                what2_dict['prodpar'] = np.double(radar.fixed_angle['data'])
            elif scan_type.upper() == 'RHI':
                # Correct for Index to scalar error - only 1 azimuth -
                # create array
                if n_datasets <= 1:
                    what2_dict['prodpar'] = np.repeat(
                        np.double(radar.fixed_angle['data']), n_datasets)
                else:
                    what2_dict['prodpar'] = np.double(
                        radar.fixed_angle['data'])
            else:
                warn("Scan type "+scan_type+" not yet supported")
        else:
            warn("Scan type "+scan_type+" not yet supported")

        #time
        _time = radar.time['data']
        start_sweep_ind = radar.sweep_start_ray_index['data']
        end_sweep_ind = radar.sweep_end_ray_index['data']
        startdate = []
        starttime = []
        enddate = []
        endtime = []
        for i, start_ind in enumerate(start_sweep_ind):
            _startdate = odim_datetime_struct + datetime.timedelta(
                seconds=_time[start_ind])
            _startdate_str = datetime.datetime.strftime(_startdate, "%Y%m%d")
            _starttime_str = datetime.datetime.strftime(_startdate, "%H%M%S")
            startdate.append(_startdate_str)
            starttime.append(_starttime_str)
            _enddate = odim_datetime_struct + datetime.timedelta(
                seconds=_time[end_sweep_ind[i]])
            _enddate_str = datetime.datetime.strftime(_enddate, "%Y%m%d")
            _endtime_str = datetime.datetime.strftime(_enddate, "%H%M%S")
            enddate.append(_enddate_str)
            endtime.append(_endtime_str)

        what2_dict['startdate'] = startdate
        what2_dict['starttime'] = starttime
        what2_dict['enddate'] = enddate
        what2_dict['endtime'] = endtime

        #fill the what2 group attributes
        ind = 0
        for i in what2_grps:
            for name in what_var_dataset:
                if what2_dict[name] is not None:
                    _create_odim_h5_attr(i, name, what2_dict[name][ind])
            ind = ind + 1

        #Dataset specific how variables
        #Necessary?
        ssri = radar.sweep_start_ray_index['data']
        seri = radar.sweep_end_ray_index['data']

        if odim_object in ['PVOL', 'AZIM', 'SCAN']:
            how2_dict = {}
            #PPI
            how_var_dataset = ['elangles', 'startazT', 'stopazT']

            el_angle = []
            T_start_az = []
            T_stop_az = []

            for i in range(n_datasets):
                el_angle_tmp = radar.elevation['data'][ssri[i]:seri[i]+1]
                el_angle.append(el_angle_tmp)
                T_start_az_tmp = (
                    radar.time['data'][ssri[i]:seri[i]+1] -
                    radar.time['data'][0])
                T_stop_az_tmp = (
                    radar.time['data'][ssri[i]:seri[i]+1] +
                    radar.time['data'][0])
                t1_epo = []
                for sec in T_start_az_tmp:
                    t_tmp = odim_datetime_struct + datetime.timedelta(
                        seconds=sec)
                    t_epo_sec = _get_sec_since_epoch(t_tmp)
                    t1_epo.append(t_epo_sec)
                T_start_az.append(t1_epo)
                t2_epo = []
                for sec in T_stop_az_tmp:
                    t_tmp = odim_datetime_struct + datetime.timedelta(
                        seconds=sec)
                    t_epo_sec = _get_sec_since_epoch(t_tmp)
                    t2_epo.append(t_epo_sec)
                T_stop_az.append(t2_epo)

            how2_dict['elangles'] = np.double(el_angle)
            how2_dict['startazT'] = np.double(T_start_az)
            how2_dict['stopazT'] = np.double(T_stop_az)

            if radar.azimuth['data'] is not None:
                how_var_dataset.extend(['startazA', 'stopazA'])
                start_azA = []
                stop_azA = []
                for i in range(n_datasets):
                    start_azA_tmp = (
                        radar.azimuth['data'][ssri[i]:seri[i]+1] -
                        radar.ray_angle_res['data'][i] / 2.)
                    start_azA_tmp[start_azA_tmp < 0.] += 360. #[0 360]
                    stop_azA_tmp = (
                        radar.azimuth['data'][ssri[i]:seri[i]+1] +
                        radar.ray_angle_res['data'][i] / 2.)
                    stop_azA_tmp[stop_azA_tmp > 360.] -= 360. #[0 360]
                    start_azA.append(start_azA_tmp)
                    stop_azA.append(stop_azA_tmp)

                how2_dict['startazA'] = np.double(start_azA)
                how2_dict['stopazA'] = np.double(stop_azA)

        # ELEV - RHI
        # how_var_dataset = ['startelA', 'stopelA', 'startelT', 'stopelT']
        # Not supported by ODIM reader yet

            #fill the how2 group attributes
            ind = 0
            for i in how2_grps:
                for name in how_var_dataset:
                    if how2_dict[name] is not None:
                        _create_odim_h5_attr(i, name, how2_dict[name][ind])
                ind = ind + 1

        # create level 3 data and what group structure and fill data
        what3_grps = []
        what3_dict = _tree()

        for i in range(n_datasets):
            what3_grps.append([])
            for j, field_name in enumerate(field_names):
                what3_id = _create_odim_h5_sub_grp(
                    datatype_grps[i][j], 'what')
                what3_grps[i].append(what3_id)

                # ODIM field name
                radar_quantity = _map_radar_quantity(field_name)
                what3_dict[i][j]['quantity'] = radar_quantity

                # get data
                data_dict = _get_data_from_field(
                    radar, i, field_name, physical=physical)

                # write data
                what3_dict[i][j]['gain'] = data_dict['gain']
                what3_dict[i][j]['offset'] = data_dict['offset']

                if data_dict['nodata'] is not None:
                    what3_dict[i][j]['nodata'] = data_dict['nodata']
                if data_dict['undetect'] is not None:
                    what3_dict[i][j]['undetect'] = data_dict['undetect']

                _create_odim_h5_dataset(
                    datatype_grps[i][j], 'data', data_dict['data'],
                    make_legend=False, compression=compression,
                    compression_opts=compression_opts)

                # Add legend data if 'label' key is present in radar quantity
                # dictionary
                if ('labels' in radar.fields[field_name] and
                        'ticks' in radar.fields[field_name]):
                    labs = radar.fields[field_name]['labels']
                    tics = radar.fields[field_name]['ticks']
                    data_legend = []
                    for lab_tic in zip(labs, tics):
                        data_legend.append(lab_tic)
                    _create_odim_h5_dataset(
                        datatype_grps[i][j], 'legend', data_legend,
                        make_legend=True, compression=compression,
                        compression_opts=compression_opts)

        # fill the what3 group attributes of data
        for i in range(n_datasets):
            for j in range(n_datatypes):
                for name in what_var_data:
                    if name in what3_dict[i][j]:
                        rq_name = what3_dict[i][j][name]
                        _create_odim_h5_attr(what3_grps[i][j], name, rq_name)
                    # else:
                    #    warn("Attribute "+name+" not specified")

    #close HDF file
    hdf_id.close()
    _check_file_exists(filename)


def _to_str(text):
    """
    Converter: From byte arrays to string if necessary.

    Parameters:
    -----------
    text: byte array
        Input for conversion

    Returns:
    -------
    text : str
        String of input array

    """
    if hasattr(text, 'decode'):
        return text.decode('utf-8')

    return text


def _check_file_exists(filename):
    """
    Check for ODIM h5 file existence

    Parameters:
    -----------
    filename : str
        Filename string pointing to ODIM h5 file that should be present.

    """
    myfile = Path(filename)
    if myfile.is_file():
        print('ODIM h5 file '+filename+' written.')
    else:
        print('Warning: ODIM h5 file '+filename+' not written.')


def _get_sec_since_epoch(time_f):
    """
    Calculate seconds since 1970-01-01 epoch with microseconds precision

    Parameters:
    -----------
    time_f : datetime
        Datetime object to process

    Returns:
    --------
    sec_since_epoch : double
        Seconds since 1970 epoch

    """
    if hasattr(time_f, 'microsecond'):
        sec_since_epoch = (
            calendar.timegm(time_f.timetuple()) +
            time_f.microsecond/1000000.0)
    else:
        sec_since_epoch = calendar.timegm(time_f.timetuple())

    return sec_since_epoch


def _tree():
    """
    Initialize a tree structure for a multidimensional dictionary

    """
    return defaultdict(_tree)


def _create_odim_h5_file(filename, access_mode='w', driver=None):
    """
    Initialize HDF5 file with h5py (https://www.h5py.org/) to write ODIM
    compliant data structure.

    Parameters:
    -----------
    filename : str
        Filename of ODIM HDF5 file to create
    access_mode : str
        'w' for going into write model
    driver : str, optional
        Optional driver specification in h5py

    Returns:
    --------
    file_id : object
        h5py file ID object

    """
    # Check that h5py module is available
    if not _H5PY_AVAILABLE:
        raise MissingOptionalDependency(
            "h5py is required to use the ODIM HDF5 writer, '" +
            "'but is not installed")

    file_id = h5py.File(filename, access_mode, driver)

    return file_id


def _create_odim_h5_grp(file_id, grp_name):
    """
    Create HDF5 group for a specified file ID.

    Parameters:
    -----------
    file_id : object
        File ID object pointing to h5py file where group should be created.
    grp_name : str
        Group name to be created

    Returns:
    --------
    grp_id : obejct
        h5py group ID object

    """
    grp_id = file_id.create_group(grp_name)

    return grp_id


def _create_odim_h5_sub_grp(grp_id, sub_grp_name):
    """
    Create HDF5 subgroup within a specified group.

    Parameters:
    -----------
    grp_id : object
        Group ID object pointing to h5py group where subgroup should be
        created.
    sub_grp_name : str
        Subgroup name to be created

    Returns:
    --------
    sub_grp_id : obejct
        h5py subgroup ID object

    """
    sub_grp_id = grp_id.create_group(sub_grp_name)

    return sub_grp_id


def _create_odim_h5_attr(grp_id, name, data):
    """
    Create and fill group (subgroup) attributes with metadata.

    Parameters:
    -----------
    grp_id : object
        h5py object where attributes shoul be created.
    name : str
        Name of the attribute to be created.
    data : data (str, int, double, array)
        Data to be saved in h5py group attribute

    """
    #Check if data is str
    if isinstance(data, str):
        string_dt = h5py.special_dtype(vlen=str)
        grp_id.attrs.create(
            name, np.array(data, dtype=string_dt), dtype=string_dt)
    else:
        grp_id.attrs.create(name, data)


def _create_odim_h5_dataset(ID, name, data_arr, make_legend=False,
                            compression="gzip", compression_opts=6):
    """
    Create and save radar field data array to h5py dataset.

    Parameters:
    -----------
    ID : object
        h5py object ID pointing to dataset to be created and filled.
    name : str
        Name of the dataset to be created
    data_arr : array
        Array containing the data
    make_legend : Bool
        If true a legend is created
    compression : str
        compression type and level. Can be "gzip", "lzf"
    compression_opts : int
        if "gzip" is choosen, the level of compression. ODIM recommends
        levels between 1 and 6
    """
    if make_legend:
        dt_tmp = [
            ('label', h5py.special_dtype(vlen=str)), ('tick', np.float)]
        ds = ID.create_dataset(
            name, (1, len(data_arr)), dtype=dt_tmp, compression=compression,
            compression_opts=compression_opts)
        data_np_arr = np.array(data_arr, dtype=dt_tmp)
        ds[0] = data_np_arr
    else:
        ID.create_dataset(
            name, data=data_arr, compression=compression,
            compression_opts=compression_opts)


def _map_radar_quantity(field_name):
    """
    Map radar field quantities to ODIM compliant quantities.

    Parameters:
    -----------
    field_name : str
        the  radar field names

    Returns:
    --------
    ODIM_field_name : str
        ODIM compliant radar field name

    """
    odim_quantity_dict = {
        'unfiltered_reflectivity': 'TH',
        'unfiltered_reflectivity_vv': 'TV',
        'reflectivity': 'DBZH',
        'reflectivity_vv': 'DBZV',
        'corrected_reflectivity': 'DBZHC',  # Non standard ODIM
        'corrected_reflectivity_vv': 'DBZVC',  # Non standard ODIM
        'corrected_unfiltered_reflectivity': 'THC',  # Non standard ODIM
        'corrected_unfiltered_reflectivity_vv': 'TVC',  # Non standard ODIM
        'reflectivity_bias': 'ZBIAS',  # Non standard ODIM
        'volumetric_reflectivity': 'etah',  # Non standard ODIM
        'volumetric_reflectivity_vv': 'etav',  # Non standard ODIM
        'radar_cross_section_hh': 'RCSH',  # Non standard ODIM
        'radar_cross_section_vv': 'RCSV', # Non standard ODIM
        'differential_reflectivity': 'ZDR',
        'unfiltered_differential_reflectivity': 'ZDRU',  # Non standard ODIM
        'corrected_differential_reflectivity': 'ZDRC',  # Non standard ODIM
        'corrected_unfiltered_differential_reflectivity': 'ZDRUC',  # Non standard ODIM
        'differential_reflectivity_in_precipitation': 'ZDRPREC', # Non standard ODIM
        'differential_reflectivity_in_snow': 'ZDRSNOW', # Non standard ODIM
        'signal_power_hh': 'DBMH',  # Non standard ODIM
        'signal_power_vv': 'DBMV',  # Non standard ODIM
        'noisedBZ_hh': 'NDBZH', # Non standard ODIM
        'noisedBZ_vv': 'NDBZV', # Non standard ODIM
        'sun_hit_power_h': 'DBM_SUNHIT', # Non standard ODIM
        'sun_hit_power_v': 'DBMV_SUNHIT', # Non standard ODIM
        'sun_hit_differential_reflectivity': 'ZDR_SUNHIT', # Non standard ODIM
        'sun_est_power_h': 'DBM_SUNEST',  # Non standard ODIM
        'sun_est_power_v': 'DBMV_SUNEST',  # Non standard ODIM
        'sun_est_differential_reflectivity': 'ZDR_SUNEST', # Non standard ODIM
        'sun_hit_h': 'POSH_SUNHIT', # Non standard ODIM
        'sun_hit_v': 'POSV_SUNHIT', # Non standard ODIM
        'sun_hit_zdr': 'POSZDR_SUNHIT', # Non standard ODIM
        'cross_correlation_ratio': 'RHOHV',
        'uncorrected_cross_correlation_ratio': 'URHOHV', # Non standard ODIM
        'corrected_cross_correlation_ratio': 'RHOHVC',  # Non standard ODIM
        'cross_correlation_ratio_in_rain': 'RHOHVRAIN', # Non standard ODIM
        'logarithmic_cross_correlation_ratio': 'LRHOHV', # Non standard ODIM
        'circular_depolarization_ratio': 'CDR',  # Non standard ODIM
        'linear_polarization_ratio': 'LDR',
        'differential_phase': 'PHIDP',
        'uncorrected_differential_phase': 'UPHIDP', # Non standard ODIM
        'corrected_differential_phase': 'PHIDPC',  # Non standard ODIM
        'system_differential_phase': 'PHIDP0',  # Non standard ODIM
        'first_gate_differential_phase': 'PHIDP0_BIN',  # Non standard ODIM
        'specific_differential_phase': 'KDP',
        'corrected_specific_differential_phase': 'KDPC',  # Non standard ODIM
        'normalized_coherent_power': 'SQIH',
        'normalized_coherent_power_vv': 'SQIV',
        'signal_to_noise_ratio_hh': 'SNRH',
        'signal_to_noise_ratio_vv': 'SNRV',
        'clutter_correction_hh': 'CCORH',  # Not used in Pyrad
        'clutter_correction_vv': 'CCORV', # Not used in Pyrad
        'radar_estimated_rain_rate': 'RATE',
        'uncorrected_rain_rate': 'URATE', # Not used in Pyrad
        'hail_intensity': 'HI',  # Not used in Pyrad
        'hail_probability': 'HP', # Not used in Pyrad
        'accumulated_precipitation': 'ACRR', # Not used in Pyrad
        'echotop_height': 'HGHT', # Not used in Pyrad
        'vertical_integrated_liquid_water': 'VIL', # Not used in Pyrad
        'velocity': 'VRADH',
        'velocity_vv': 'VRADV',
        'corrected_velocity': 'VRADHC',  # Non standard ODIM
        'dealiased_velocity': 'VRADDH',
        'dealiased_corrected_velocity': 'VRADDHC',  # Non standard ODIM
        'dealiased_velocity_vv': 'VRADDV',
        'retrieved_velocity': 'VRADEST',  # Non standard ODIM
        'retrieved_velocity_std': 'sd_vvp',  # special vol2bird
        'velocity_difference': 'VDIFF',  # Non standard ODIM
        'spectrum_width': 'WRADH',
        'corrected_spectrum_width': 'WRADHC',  # Non standard ODIM
        'spectrum_width_vv': 'WRADV',
        'eastward_wind_component': 'UWND',
        'northward_wind_component': 'VWND',
        'azimuthal_horizontal_wind_component': 'AHWND',  # Non standard ODIM
        'vertical_wind_component': 'w', # Standard for vertical profile
        'radial_wind_shear': 'RSHR',  # Not used in Pyrad
        'azimuthal_wind_shear': 'ASHR',  # Not used in Pyrad
        'range_azimuthal_wind_shear': 'CSHR',  # Not used in Pyrad
        'elevation_wind_shear': 'ESHR',  # Not used in Pyrad
        'range_elevation_wind_shear': 'OSHR',  # Not used in Pyrad
        'horizontal_wind_shear': 'HSHR',  # Not used in Pyrad
        'vertical_wind_shear': 'VSHR',
        'three_dimensional_shear': 'TSHR', # Not used in Pyrad
        'wind_speed': 'ff', # Standard for vertical profile
        'wind_direction': 'dd',  # Standard for vertical profile
        'specific_attenuation': 'AH',  # Non standard ODIM
        'corrected_specific_attenuation': 'AHC',  # Non standard ODIM
        'path_integrated_attenuation': 'PIA',  # Non standard ODIM
        'corrected_path_integrated_attenuation': 'PIAC',  # Non standard ODIM
        'specific_differential_attenuation': 'ADP',  # Non standard ODIM
        'corrected_specific_differential_attenuation': 'ADPC',  # Non standard ODIM
        'path_integrated_differential_attenuation': 'PIDA',  # Non standard ODIM
        'corrected_path_integrated_differential_attenuation': 'PIDAC',  # Non standard ODIM
        'temperature': 'TEMP',  # Non standard ODIM
        'iso0': 'ISO0',  # Non standard ODIM
        'height_over_iso0': 'HISO0',  # Non standard ODIM
        'cosmo_index': 'COSMOIND',  # Non standard ODIM
        'hzt_index': 'HZTIND',  # Non standard ODIM
        'melting_layer': 'ML',  # Non standard ODIM
        'visibility': 'VIS',  # Non standard ODIM
        'radar_echo_id': 'ECHOID',  # Non standard ODIM
        'clutter_exit_code': 'CLT',  # Non standard ODIM
        'occurrence': 'OCC',  # Non standard ODIM
        'frequency_of_occurrence': 'OCCFREQ',  # Non standard ODIM
        'radar_border': 'BRDR',  # Not used in Pyrad
        'signal_quality_index': 'QUIND',
        'radar_echo_classification': 'CLASS',
        'hydroclass_entropy': 'ENTROPY',  # Non standard ODIM
        'proportion_AG': 'propAG',  # Non standard ODIM
        'proportion_CR': 'propCR',  # Non standard ODIM
        'proportion_LR': 'propLR',  # Non standard ODIM
        'proportion_RP': 'propRP',  # Non standard ODIM
        'proportion_RN': 'propRN',  # Non standard ODIM
        'proportion_VI': 'propVI',  # Non standard ODIM
        'proportion_WS': 'propWS',  # Non standard ODIM
        'proportion_MH': 'propMH',  # Non standard ODIM
        'proportion_IH': 'propIH',  # Non standard ODIM
        'time_avg_flag': 'TAFLAG',  # Non standard ODIM
        'colocated_gates': 'COLGATES',  # Non standard ODIM
        'number_of_samples': 'ns',  # standard vertical profile
        'bird_density': 'dens',  # standard vol2bird
        'standard_deviation': 'STD',  # Non standard ODIM
        'sum': 'SUM',  # Non standard ODIM
        'sum_squared': 'SUM2',  # Non standard ODIM
        'height_resolution': 'width',  # Special vol2bird
        'gap': 'gap',  # Special vol2bird
        'bird_reflectivity': 'eta',  # Special vol2bird
        'number_of_samples_velocity': 'n',  # Special vol2bird
        'number_of_samples_reflectivity': 'n_dbz',  # Special vol2bird
        'number_of_samples_velocity_all': 'n_all',  # Special vol2bird
        'number_of_samples_reflectivity_all': 'n_dbz_all'  # Special vol2bird
    }
    if field_name not in odim_quantity_dict:
        warn('Field name '+field_name+' does not have an ODIM quantity ' +
             'equivalent. The same field name will be used.')
        return field_name

    return odim_quantity_dict[field_name]


def _get_data_from_field(radar, sweep_ind, field_name, physical=True):
    """
    Extract data from radar field object with respect to different datasets
    and datatypes.

    Parameters:
    -----------
    radar : object
        Radar object
    sweep_ind : int
        sweep index, equal to ODIM dataset index
    field_name : str
        name of field to convert
    physical : bool
        If true the data will be stored in physical units. Otherwise it will
        be converted in binary

    Returns:
    --------
    data_dict : dict
        Dictionary containing the data, gain, offset, nodata and undetect

    """
    data_ph = np.ma.asarray(radar.get_field(sweep_ind, field_name, copy=True))

    if physical:
        fill_value = radar.fields[field_name].get(
            '_FillValue', np.double(get_fillvalue()))
        if fill_value is not None:
            data = data_ph.filled(fill_value)
            nodata = fill_value
        else:
            data = np.asarray(data_ph)
            nodata = None
        undetect = None
        gain = 1.
        offset = 0.
    else:
        dtype = 'uint8'
        nvals = 256
        if '_Write_as_dtype' in radar.fields[field_name]:
            dtype = radar.fields[field_name]['_Write_as_dtype']
            if dtype not in ('uint8', 'uint16'):
                warn('Type '+dtype+' not supported. uint8 will be used')
                dtype = 'uint8'
                nvals = 256
            elif dtype == 'uint8':
                nvals = 256
            else:
                nvals = 65536

        nodata = radar.fields[field_name].get('nodata', 0)
        undetect = radar.fields[field_name].get('undetect', None)

        reserved = 0
        if nodata is not None:
            reserved = 1
        if undetect is not None:
            reserved += 1

        shift_up = 0
        if nodata == 0 or undetect == 0:
            shift_up = 1
        if nodata == 1 or undetect == 1:
            shift_up = 2

        if (('scale_factor' in radar.fields[field_name]) and
                ('add_offset' in radar.fields[field_name])):
            gain = radar.fields[field_name]['scale_factor']
            offset = radar.fields[field_name]['add_offset']

            shift_down = 0
            if nodata == nvals-1 or undetect == nvals-1:
                shift_down = 1
            if nodata == nvals-2 or undetect == nvals-2:
                shift_down = 2
            val_min = gain*shift_up+offset
            val_max = gain*(nvals-1-shift_down)+offset

            # Check if data within range
            ind = np.ma.where(data_ph < val_min)
            if ind[0].size > 0:
                warn(str(ind[0].size)+' data points below the minimum value ' +
                     str(val_min))
                data_ph[ind] = val_min
            ind = np.ma.where(data_ph > val_max)
            if ind[0].size > 0:
                warn(str(ind[0].size)+' data points above the maximum value ' +
                     str(val_max))
                print(data_ph[ind])
                data_ph[ind] = val_max
        else:
            val_min = np.ma.min(data_ph)
            val_max = np.ma.max(data_ph)
            nsteps = nvals-reserved
            gain = (val_max-val_min)/nsteps
            offset = val_min-shift_up*gain

            warn('gain and offset not specified. Optimal scale with gain ' +
                 str(gain)+' and offset '+str(offset)+' will be used')

        data = (data_ph-offset)/gain
        if nodata is not None:
            data = data.filled(nodata)
        data = np.asarray(data, dtype=eval('np.'+dtype))

    data_dict = {
        'data': data,
        'gain': gain,
        'offset': offset,
        'nodata': nodata,
        'undetect': undetect
    }
    return data_dict


def _map_radar_to_how_dict(radar_obj):
    """
    Tries to map data in a radar sub object (e.g. radar.instrument_parameters)
    to ODIM how attributes.

    Parameters:
    -----------
    radar_obj : Radar
        Containing a radar sub object

    Returns:
    --------
    dict_odim : dictionary
        Dictionary of ODIM how variables

    """
    # Variables in the instrument_parameter convention and
    # radar_parameters sub-convention that will be read from and written to
    # ODIM H5 files using Py-ART.
    # The meta_group attribute cannot be used to identify these parameters as
    # it is often set incorrectly.

    _INSTRUMENT_PARAMS = [
        #instrument parameters sub-convention
        'frequency',
        'pulse_width',
        'prt',
        'nyquist_velocity',
        'n_samples',
        # radar_parameters sub-convention
        'radar_antenna_gain_h',
        'radar_antenna_gain_v',
        'calibration_constant_hh',
        'calibration_constant_vv',
        'radar_beam_width_h',
        'radar_beam_width_v',
        'radar_receiver_bandwidth',
        'radar_rx_bandwidth'    # non-standard
        #'radar_measured_transmit_power_h', # not found in radar object
        #'radar_measured_transmit_power_v', # not found in radar object
        #'measured_transmit_power_v',    # not found in radar object
        #'measured_transmit_power_h'     # not found in radar object
        ]

    _RADAR_METADATA = [
        'system',
        'software',
        'sw_version'
        ]

    c = 299792458 #Speed of light

    dict_odim = {}

    for key in radar_obj.keys():
        if key in _INSTRUMENT_PARAMS:
            if key == 'frequency':
                dict_odim['wavelength'] = c/np.double(radar_obj[key]['data'])
            if key == 'pulse_width':
                dict_odim['pulsewidth'] = (
                    np.double(radar_obj[key]['data'])*1e6)
            if key == 'prt':
                dict_odim['highprf'] = 1 / np.double(radar_obj[key]['data'])
            if key == 'nyquist_velocity':
                dict_odim['NI'] = np.double(radar_obj[key]['data'])
            if key == 'n_samples':
                dict_odim['Vsamples'] = np.int64(radar_obj[key]['data'])
            if key == 'radar_antenna_gain_h':
                dict_odim['antgainH'] = np.double(radar_obj[key]['data'])
            if key == 'radar_antenna_gain_v':
                dict_odim['antgainV'] = np.double(radar_obj[key]['data'])
            if key == 'calibration_constant_hh':
                dict_odim['radconstH'] = np.double(radar_obj[key]['data'])
            if key == 'calibration_constant_vv':
                dict_odim['radconstV'] = np.double(radar_obj[key]['data'])
            if key == 'radar_beam_width_h':
                dict_odim['beamwH'] = np.double(radar_obj[key]['data'])
            if key == 'radar_beam_width_v':
                dict_odim['beamwV'] = np.double(radar_obj[key]['data'])
            if key in ['radar_receiver_bandwidth', 'radar_rx_bandwidth']:
                dict_odim['RXbandwidth'] = np.double(radar_obj[key]['data'])
        elif key in _RADAR_METADATA:
            dict_odim[key] = _to_str(radar_obj[key])
        else:
            warn("Unknown how parameter: %s, "%(key)+"not written to file.")

    return dict_odim
